# -*- coding: utf-8 -*-
# @File    :   base.py
# @Time    :   2023/11/08 13:50:38
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
'''


####################### imports ###########################

from abc import ABC, abstractmethod

class BaseEvalutor(ABC):
    def __init__(self, config):
        self.config = config

    @abstractmethod
    def normalize_sample(self, sample):
        """ standardize the sample to a dict with keys 
            sample_id

        IMPORTANT: values should not be dict, or it will be extremely slow
        """
        return sample

    @abstractmethod
    def evaluate(self):
        """ 

        """
        raise NotImplementedError

    
 



import os 
import faiss
from sentence_transformers import SentenceTransformer 

from .base_retriever import Retriever

class VectorRetriever(Retriever):
    def __init__(self, data_source, key=None):
        self.data_source = data_source
        assert key is not None, "key is not specified"
        if type(key) is str:
            self.items4retrieval = [ sample[key] for sample in data_source]
        elif callable(key):
            # 以构建更多样的检索形式，而非简单的使用sample[key]作为检索项
            self.items4retrieval = [ key(sample) for sample in data_source]
        else:
            raise Exception(f"key should be dict key or callable function")
        self.key = key 
        self.model = self._init_encoding_model()
        self._build_index()
    

    def _init_encoding_model(self):
        model = SentenceTransformer("/sshfs/pretrains/sentence-transformers/all-MiniLM-L6-v2")
        return model 
            
    def _build_index(self):
        cache_file = f"{ self.key if type(self.key) is str else self.key.__name__ }_index.bin"

        if os.path.exists(cache_file):
            self.index = faiss.read_index(cache_file)
            print(f"index loaded from {cache_file}")
        else:
            self.index = faiss.IndexFlatIP(384)
            vectors = self.model.encode(self.items4retrieval, batch_size=128, show_progress_bar=True)
            self.index.add(vectors)
            faiss.write_index(self.index, cache_file)
            print(f"index built and saved to {cache_file}")

    
    def get_topk_samples(self, query_vector, topk=5):
        """ 将输入query分词，返回self.topk个 sample """
     
        D, I = self.index.search(query_vector.reshape(1,-1), topk)
        topk_samples = [ self.data_source[i] for i in I[0]]
        return topk_samples


# -*- coding: utf-8 -*-
# @File    :   metric_fns.py
# @Time    :   2023/11/20 14:05:24
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    用于评估 GQL 生成模型的指标函数
'''
from qdls.reg.register import Register
from qdls.gql.cypher.utils.syntax import syntax_check as syntax_check_cypher
from qdls.gql.sparql.utils.syntax import syntax_check as syntax_check_sparql

from qdls.gql.cypher.utils.kqa_eval import exec_one_sample as exec_one_sample_cypher
from torchmetrics.functional import sacre_bleu_score

from pygments.lexers import get_lexer_by_name
from pygments.lexers.rdf import SparqlLexer
from pygments.token import *

sp_lexer = SparqlLexer()
cy_lexer = get_lexer_by_name('py2neo.cypher') # pip install py2neo to fix

metric_fns = Register('metric_fns')

# from .utils import detect_query_language, postprocess_sparql_tokens

@metric_fns.register("exact_match")
def match_query(pred, gold, lexer='cypher', verbose=False):
    """ 将cypher or sparql 语言解析为语义单元，比较是否匹配
        如果出现ERROR tag 则表示生成的语法错误， return 0 
    """
    lexer = sp_lexer if lexer == 'sparql' else cy_lexer
    pred_units = []
    for tag, substr in lexer.get_tokens(pred):
        if tag is Token.Error:
            return 0
        # print(tag, substr)
        if substr.strip() != "":
            pred_units.append(substr)
    gold_units = [ substr for tag, substr in lexer.get_tokens(gold) if substr.strip() != "" ]
    res = (gold_units == pred_units)
    if res:
        return 1
    else:
        if verbose:
            print(f'{" ".join(gold_units)} \n {" ".join(pred_units)}')
        return 0
    

def executable_fn_cypher(pred):
    """ 判断预测的是否可执行 """
    flag, tree, parser = syntax_check_cypher(pred)
    return 1.0 if flag else 0.0


def executable_fn_sparql(pred):
    """ 判断预测的是否可执行 """
    flag, tree, parser = syntax_check_sparql(pred)
    return 1.0 if flag else 0.0

@metric_fns.register("executable")
def executable_fn(pred, lang):
    if lang == 'sparql':
        return executable_fn_sparql(pred)
    elif lang in ['cql', 'cypher']:
        return executable_fn_cypher(pred) 
    else:
        raise Exception(f"{lang} is not supported!")


@metric_fns.register("bleu")
def bleu_fn(pred, refs):
    """ 默认是用 mteval-v13a 进行分词 return bleu-4 """
    return sacre_bleu_score([pred], [refs], tokenize='char').item()



def calc_metrics_per_sample(sample, ref_key, metrics=None, neo4j_config=None):
    """ 默认的预测 key 为 pred 
        sample normalized_sample()的结果
        ref_key 为参考答案的key: cypher,cql, sparql
        metrics 为要计算的指标， 默认为 ['bleu', 'executable', 'exact_match']
        neo4j_config 为neo4j的配置 
    """
    if metrics is None:
        metrics = ['bleu', 'executable', 'exact_match']
    else:
        assert isinstance(metrics, list), f"metrics should be a list of str, but got f{metrics}"

    if 'bleu' in metrics:
        bleu = bleu_fn(sample['pred'], [sample[ref_key]])
        sample['bleu'] = bleu
    
    if 'executable' in metrics:
        sample['executable'] = executable_fn(sample['pred'], ref_key)

    if 'exact_match' in metrics:
        sample['exact_match'] = match_query(sample['pred'], sample[ref_key], lexer=ref_key)

    if 'is_correct' in metrics or 'exec_info' in metrics:
        assert ref_key in ['cql', 'cypher'], f"only cql or cypher is supported, but got {ref_key}"
        is_correct, info = exec_one_sample_cypher(sample, neo4j_config, key='pred')
        sample['is_correct'] = is_correct
        sample['exec_info'] = info

    # import pdb;pdb.set_trace();
    return sample

def nltk_bleu():
    import nltk
    reference = ['今天天气很好吗']

    # 候选翻译（机器翻译结果）
    candidate = '今日天气很好。'

    # 将参考翻译和候选翻译转换为标记化的句子列表
    reference_tokens = nltk.word_tokenize(reference[0], language='chinese')
    candidate_tokens = nltk.word_tokenize(candidate, language='chinese')

    # 计算BLEU分数
    bleu_score = nltk.translate.bleu_score.sentence_bleu([reference_tokens], candidate_tokens)

    print("BLEU score:", bleu_score)

if __name__ == '__main__':
    # print(bleu_fn('hello world', ['hello world']))
    nltk_bleu()


import os,sys
PROJ_DIR="/Users/qing/Library/Mobile Documents/com~apple~CloudDocs/workspace/Code/qdls/src"
# sys.path.append(PROJ_DIR)
sys.path.insert(0, PROJ_DIR)


# from src.qdls.reg import registry 
# import qdls
# print(dir(qdls.__all__))

from qdls.reg.register import registers, import_all_modules_for_register
print(registers.model._dict)
import_all_modules_for_register( only_data=False)
print(registers.model._dict)

print(registers.collator._dict)
print(registers.process_function._dict)
print("data module", registers.datamodule._dict)
# import pdb;pdb.set_trace();

@registers.process_function.register("my_test")
def my_test():
    print("This is my test function.")
    return 

from qdls.utils import print_string

registers.process_function.get("my_test")()
print_string(registers.process_function.keys())


print_string("This is direct import.")

from qdls.reg.process_functions.kqa import kqa_test_function

kqa_test_function()
 


# -*- coding: utf-8 -*-
# @File    :   kqa_eval.py
# @Time    :   2023/06/26 16:36:09
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    elvauate KQA Cypher
'''

import os 
import neo4j
from neo4j import GraphDatabase, Query

from qdls.data import load_json, save_json
from qdls.gql.cypher.utils.syntax import syntax_check

from concurrent.futures import ThreadPoolExecutor
from multiprocessing import Pool
from argparse import Namespace
from tqdm import tqdm

from qdls.utils import print_string

units = ['metre', 'pound', 'minute', 'United States dollar', 'mile', '1', 'centimetre', 'square metre', 'square kilometre', 
'hour', 'second', 'kilogram', 'inch', 'hertz', 'foot', 'Japanese yen', 'year', 'square mile', 'kilometre', 'act', 'audio track',
 'millimetre', 'Danish krone', 'hectare', 'euro', 'Argentine peso', 'beats per minute', 'Nigerian naira', 'barrels per day', 
 'kilogram per cubic metre', 'kilometre per hour', 'month', 'annum', 'Singapore dollar', 'cubic metre per second', 'Czech koruna', 
 'day', 'pound sterling', 'percentage', 'acre', 'workforce', 'square foot', 'South Korean won', 'Canadian dollar', 'Indian rupee', 
 'reservist', 'volt', 'Russian ruble', 'active duty military personnel', 'son', 'atomic mass unit', 'Hong Kong dollar', 
 'years old', 'drawing', 'book', 'gram per cubic centimetre', 'painting', 'gram per kilogram', 'sculpture', 'degree Celsius', 
 'cubic hectometre', 'photograph', 'yottagram', 'gram per mole', 'geographic map', 'astronomical unit', 'Australian dollar', 
 'degree', 'Philippine peso', 'chapter', 'Deutsche Mark', 'cubic kilometre', 'gigawatt', 'solar radius', 'kelvin', 
 'disability-adjusted life year', 'solar mass', 'degree Fahrenheit', 'millimeter of mercury', 'milligram per cubic meter', 
 'volume percent', 'electronvolt', 'metre per second', 'hectopascal', 'joule per mole-kelvin', 'sol', 'daughter', 
 'kilojoule per mole', 'watt per meter kelvin', 'gigapascal', 'cubic metre per ton', 'picometre', 'amperes per volt meter', 
 'millipascal-second', 'Caribbean guilder', 'megajoule per kilogram', 'liter per kilogram', 'grams per cubic meter', 
 'Lebanese pound', 'gram per 100 gram of solvent', 'square versta', 'gram', 'tonne', 'week', 'joule per tesla', 'country', 
 'Geary–Khamis dollar', 'cases per 100000 person-years', 'European Currency Unit', 'Swiss franc', 'hryvnia', 'Swedish krona', 
 'crore', 'Thousands', 'gigabyte', 'torr', 'movie theater', 'academic term']

def serialize_result(res):
    """现在的方法比较日期有BUG，使用此函数转换后再比较  """
    if type(res) is neo4j.time.Date:
        return res.isoformat()
    elif type(res) is dict:
        return {k:serialize_result(v) for k,v in res.items()}
    elif type(res) is list:
        return [serialize_result(_) for _ in res]
    else:
        return str(res)


def exec_one_sample(sample, config, key):
    driver = GraphDatabase.driver(uri=config.neo4j_uri, auth=(config.neo4j_user, config.neo4j_passwd))

    # query = sample['pred'] if 'pred' in sample else sample['cypher']
    query = sample[key]
    # 不是很需要这个了
    # query = fix_illeal_relation(query)    # 对生成的语句进行后处理，In case `` is not generated
    if not syntax_check(query):
        return False, "syntax pre-check failed"
    ans = sample.get('answer', None)
    unit  = None                           # 答案中可能有单位
    if ans is not None:
        if query.strip()[-7:] == "type(p)" or query.strip()[-10:] == "type ( p )": #问的是关系
            if ans == "position played on team / speciality":
                ans = '_speciality'
            if " " in ans:
                ans = ans.replace(" ", "_")
            if "-" in ans or "(" in ans or "\xa0" in ans or "." in ans or "'" in ans:
                ans = f"`{ans}`"
        # elif any(u in ans for u in units):  # 答案中带有关系
        #     try:
        #         n, *unit_parts = ans.split(" ")
        #         ans = f'{float(n)}'
        #         unit = " ".join(unit_parts)
        #     except:
        #         pass 
        elif any(u in ans for u in units) and " " in ans:  # 答案中包含单位, fix_bug:有单位则须有空格
            if "1" in ans and ans.replace(" ", "").isnumeric():  # ISBN 等如 '0000 0000 4361 9306'
                pass 
            else:
                try:
                    n, *unit_parts = ans.split(" ")
                    unit = " ".join(unit_parts)
                    if unit in units:
                        ans = f'{float(n)}'
                except:
                    # print(f"execution answer units converting failed: {ans}, {[ u for u in units if u in ans]}")
                    pass 
    # 开始执行
    matched = None
    with driver.session(database='neo4j') as session:
        try:
            q = Query(query, timeout=config.timeout)
            res = session.run(q).data()

        except neo4j.exceptions.CypherSyntaxError as e:
            # print(query, " ||| ", e)
            res = "syntax: " + query 
            return False,  res
        except neo4j.exceptions.DatabaseError as e:
            return False, "db error incomplete: "+ query
        
        result_str = str(serialize_result(res))
        
        if len(res) == 0:
            if ans == 'no':
                matched = True 
            else:
                # return "no result: " + query, matched
                matched = False
        elif ans is None:
            # 没有提供答案，只返回查询结果
            return  None, res 
        else:
            # res  = str(res[0]) + "query: "+ query
            if ans in result_str:
                if unit is not None and unit not in result_str:
                    matched = False
                else:
                    matched = True
            elif ans == 'no' and ('False' in result_str or "None" in result_str):  # 值节点不存在该属性名，则对应null
                matched = True 
            elif ans == 'yes' and 'True' in result_str:
                matched = True 
            else:
                matched = False
    driver.close()
    # 返回查询结果，以及是否匹配成功（结果正确）
    return matched, result_str


def process_execute(queries, nproc=8, config=None, key='cypher'):
    
    assert config is not None, f"config is None, please set config first"

    Results = [] 
    if nproc == 1:
        print_string("sequential execution")
        for sample in tqdm(queries):
            Results.append(exec_one_sample(sample, config, key))
    else:
        with Pool(nproc) as pool:
            R = {}
            for sample in queries:
                future = pool.apply_async(exec_one_sample, (sample, config, key))
                R[future] = sample
            
            for future in tqdm(R):
                try:
                    res = future.get()
                except Exception as e:
                    print(e)
                    print(R[future])
                    print("+"*80)
                    res = False, f"Neo4jError: {R[future]['pred']}"
                Results.append(res)
    assert len(Results) == len(queries), f"{len(R)} != {len(queries)}"
    return Results


def cypher_exec_eval(path, nproc=16, key='pred', resave=False, post_process_fn=None, config=None):
    """ 
        path: 可以是字符串（读取文件）也可以是读取后的对象list 
        key: 每一个sample字典中, 预测的sparql语句的key
        nproc: 并行处理的进程数
        resave: bool or string  是否重新保存结果, 保存在原目录, 文件名前加上exec_
        post_process_fn: 对每一个sample进行后处理的函数，返回一个新的sample
    """
    data = load_json(path) if type(path) is str else path 
    # queries = [ s['generated'] for s in data]
    # Results = threads_execution(data, nthreads=32)
    if post_process_fn is None:
        print_string("post_process_fn is None, make sure `key` do not need to be processed")
    if post_process_fn is not None:
        data = [ post_process_fn(s) for s in data]
    Results = process_execute(data, nproc=nproc, config=config, key=key)
    
    is_right = [ 1 for match, res_str in Results if match is True ]
    acc = 100 * len(is_right)/ len(Results)
    print(f"{len(is_right)} of {len(Results)} is correct, {acc:.3f}")

    if resave:
        if type(resave) is not str:
            assert type(path) is str, "resave=True, path should be a string, or set resave=filepath"
        else:
            path = resave

        for sample, exec_res in zip(data, Results):
            sample['exec'] = exec_res
        print_string(f"resaving to {resave}")
        new_filename = "exec_" + os.path.basename(resave)
        new_path = os.path.join(os.path.dirname(resave), new_filename)
        save_json(data, new_path)

    return acc, Results



if __name__ == '__main__':

    config = Namespace(neo4j_uri="neo4j://127.0.0.1:28892", neo4j_user="neo4j", neo4j_passwd="kqa", timeout=10)
    pass 




from qdls.gql.cypher.utils.kqa_eval import cypher_exec_eval
from argparse import Namespace

# path = "/Users/qing/Downloads/sampled_50.json"

# config = Namespace(neo4j_uri="neo4j://172.23.148.83:28892", neo4j_user="neo4j", neo4j_passwd="kqa", timeout=10)
# cypher_exec_eval(path, key='cypher', nproc=1, config=config)

#############################################

# 需要先修改 rdflib 
from qdls.gql.sparql.utils.kqa_eval import sparql_exec_acc, DataForSPARQL


virtuoso_address = "http://172.23.148.83:28890/sparql"
virtuoso_graph_uri = 'kqaspcy'
config = Namespace(
    virtuoso_address = virtuoso_address,
    virtuoso_graph_uri=virtuoso_graph_uri,
)

kb = DataForSPARQL("/Users/qing/Downloads/kb.json")
# sparql_exec_acc(path, key='sparql', nproc=1, config=config)
path = '/Users/qing/workspace_local/paper_working/Sp2cy_ACL2023/data/graph pattern/val.json'
sparql_exec_acc(path, key='sparql', nproc=1, config=config, kb=kb)



import jieba 
from rank_bm25 import BM25Okapi
from qdls.gql.cypher.utils.parse import split_query # TODO: unified interface for cypher sparql, etc

from .base_retriever import Retriever

class BM25Retriever(Retriever):
    def __init__(self, data_source, key=None, lang='en') -> None:
        """ 
            data_source: list of samples
            key: key to retrieve
        """
        self.data_source = data_source
        assert key is not None, "key is not specified"
        if type(key) is str:
            self.items4retrieval = [ sample[key] for sample in data_source]
        elif callable(key):
            # 以构建更多样的检索形式，而非简单的使用sample[key]作为检索项
            self.items4retrieval = [ key(sample) for sample in data_source]
        else:
            raise Exception(f"key should be dict key or callable function")

        self.lang = lang
        if lang == 'en':
            tokenized_corpus = [doc.split(" ") for doc in self.items4retrieval]
        elif lang == 'zh':
            tokenized_corpus = [ list(jieba.cut(doc)) for doc in self.items4retrieval]
        elif lang == 'cypher':
            tokenized_corpus = [ split_query(doc) for doc in self.items4retrieval]
        else:
            raise ValueError(f"lang must be 'en' or 'zh', but got {lang}")
        self.bm25 = BM25Okapi(tokenized_corpus)


    def get_topk_samples(self, query, topk=5):
        """ 将输入query分词，返回self.topk个 sample """
        if self.lang == "en":
            tokenized_query = query.split(" ")
        elif self.lang == 'zh':
            tokenized_query = list(jieba.cut(query))
        elif self.lang == 'cypher':
            tokenized_query = split_query(query)
        else:
            raise ValueError(f"lang must be 'en' or 'zh', but got {self.lang}")
        topk_samples = self.bm25.get_top_n(tokenized_query, self.data_source, n=topk)
        return topk_samples
    
